import * as THREE from 'three'
import React, { useEffect, useRef, useState } from 'react'
import { SphereProps, useSphere } from '@react-three/cannon'
import { useThree, useFrame } from 'react-three-fiber'
import { Event } from 'three'

const SPEED = 5
const keys: Record<string, string> = {
  KeyW: 'forward',
  KeyS: 'backward',
  KeyA: 'left',
  KeyD: 'right',
  ShiftLeft: 'running',
  Space: 'jump',
}
const moveFieldByKey = (key: string) => keys[key]
const direction = new THREE.Vector3()
const frontVector = new THREE.Vector3()
const sideVector = new THREE.Vector3()

const usePlayerControls = () => {
  const [movement, setMovement] = useState({
    forward: false,
    backward: false,
    left: false,
    right: false,
    running: false,
    jump: false,
  })
  useEffect(() => {
    const handleKeyDown = (e: Event) => {
      setMovement((m) => ({ ...m, [moveFieldByKey(e.code)]: true }))
    }
    const handleKeyUp = (e: Event) =>
      setMovement((m) => ({ ...m, [moveFieldByKey(e.code)]: false }))
    document.addEventListener('keydown', handleKeyDown)
    document.addEventListener('keyup', handleKeyUp)
    return () => {
      document.removeEventListener('keydown', handleKeyDown)
      document.removeEventListener('keyup', handleKeyUp)
    }
  }, [])
  return movement
}

// TODO Improve physics in player
const Player = (props: SphereProps) => {
  const [ref, api] = useSphere(() => ({
    type: 'Dynamic',
    fixedRotation: true,
    position: [0, 4, 0],
    mass: 9,
    linearDamping: 0,
    angularDamping: 0,
    material: { friction: 0, restitution: 0 },
    ...props,
  }))
  const { forward, backward, left, right, running } = usePlayerControls()
  const { camera } = useThree()
  const velocity = useRef([0, 0, 0])
  useEffect(() => void api.velocity.subscribe((v) => (velocity.current = v)), [])
  useFrame(() => {
    if (!ref.current) return
    camera.position.copy(ref.current.position)
    frontVector.set(0, 0, Number(backward) - Number(forward))
    sideVector.set(Number(left) - Number(right), 0, 0)
    direction
      .subVectors(frontVector, sideVector)
      .normalize()
      .multiplyScalar(running ? SPEED * 2 : SPEED)
      .applyEuler(camera.rotation)

    if (forward || backward || left || right) {
      api.velocity.set(direction.x, -SPEED / 2, direction.z)
    } else {
      api.velocity.set(0, 0, 0)
    }

    // TODO enable jump if cheating
    //if (jump && Math.abs(parseFloat(velocity.current[1].toFixed(2))) < 0.05)
    //  api.velocity.set(velocity.current[0], 10, velocity.current[2])
  })
  return <mesh ref={ref} />
}

export default Player
